//-----------------------------------------------------------------------
//  Jisp - The Java Indexed Serialization Package
//  Version 3
//
//  JispDemo.java
//-----------------------------------------------------------------------
// 
//  Copyright (C) 1996-2003 Scott Robert Ladd
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or (at
//  your option) any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//-----------------------------------------------------------------------
//  A COMMERCIAL LICENSE IS AVAILABLE from the author; you can contact
//  him at his web site or via the following e-mail address:
//
//      coyote@coyotegulch.com
//
//  See the LICENSE.POLICY file for more information on Jisp is licensed.
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//        http://www.coyotegulch.com
//-----------------------------------------------------------------------

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import com.coyotegulch.jisp.*;

public class JispDemo extends JPanel
{
    // utility function
    private void centerDialog(JDialog dlg)
    {
        // center the dialog
        Rectangle screenRect = m_frame.getBounds();
        int centerWidth  = screenRect.width  < dlg.getSize().width  ? screenRect.x : screenRect.x + screenRect.width / 2  - dlg.getSize().width / 2;
        int centerHeight = screenRect.height < dlg.getSize().height ? screenRect.y : screenRect.y + screenRect.height / 2 - dlg.getSize().height / 2;
        dlg.setLocation(centerWidth, centerHeight);
    }
    
    // inner classes
    protected static class AppCloser extends WindowAdapter
    {
        public void windowClosing(WindowEvent e)
        {
            System.exit(0);
        }
    }

    // cancel any running threads
    private class ActionCancel extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            m_active = false;
        }
    }
    
    // build a database, in a separate thread    
    private static String INPUT_ALL_NAME   = "zipcodes.txt";
    private static String INPUT_80000_NAME = "zip80000.txt";
    private static String DATABASE_NAME    = "zip.data";
    private static String ZIP_INDEX_NAME   = "zip.index";
    private static String CITY_INDEX_NAME  = "city.index";
    
    private class BuildThread extends Thread
    {
        public void run()
        {
            // now we're running                
            m_active = true;
        
            try
            {
                // disable UI while working
                setUIEnabled(false);
                
                // clear output display
                setMessage("");
                
                // get order and bucket size
                int order   = Integer.valueOf(m_txtOrder.getText()).intValue();
                
                if (order < 5)
                {
                    order = 5;
                    m_txtOrder.setText("5");
                }
                    
                addMessage("Creating ZIP database from " + (m_chkSmallDB.isSelected() ? INPUT_80000_NAME : INPUT_ALL_NAME));
                
                // delete old files
                File killit = new File(DATABASE_NAME);
                
                if (killit.exists())
                    killit.delete();
    
                killit = new File(ZIP_INDEX_NAME);
                
                if (killit.exists())
                    killit.delete();
    
                killit = new File(CITY_INDEX_NAME);
                
                if (killit.exists())
                    killit.delete();
                
                // create database
                IndexedObjectDatabase database = new IndexedObjectDatabase(DATABASE_NAME,true);

                // open indexes and attach them to the database                
                BTreeIndex codeIndex = new BTreeIndex(ZIP_INDEX_NAME,order,false);
                BTreeIndex cityIndex = new BTreeIndex(CITY_INDEX_NAME,order,true);
                database.attachIndex(codeIndex);
                database.attachIndex(cityIndex);
                
                // flag to break out of loops if an exception is thrown
                boolean broken = false;
    
                // counter
                long counter = 0;
    
                // open text file
                BufferedReader reader = null;
                reader = new BufferedReader(new FileReader(m_chkSmallDB.isSelected() ? INPUT_80000_NAME : INPUT_ALL_NAME));
                
                // while we have data to read...
                while (m_active && reader.ready()) //  && (count < limit))
                {
                    // read a line
                    String line = reader.readLine();
    
                    if (line.length() < 1)
                    {
                        m_active = false;
                        continue;
                    }
                    
                    int i;
    
                    // extract ZIP
                    String textZIP = "";
    
                    for (i = 0; (line.charAt(i) != ','); ++i)
                        textZIP += line.charAt(i);
    
                    int zip = Integer.valueOf(textZIP).intValue();
                    
                    // extract city name
                    ++i;
    
                    String city = line.substring(i);
    
                    // find state
                    String state = FindStateByZip.findState(zip);
    
                    // count it
                    ++counter;
                    addMessage(counter + ": " + city + ", " + state + " (" + zip + ")");
    
                    // write the record (we only have one index, but we still need the key array!
                    OrderedObject [] keyArray = new OrderedObject[2];
                    keyArray[0] = new IntKey(zip);
                    keyArray[1] = new StringKey(city.trim().toUpperCase() + state.trim().toUpperCase()); 
    
                    DBRecord record = new DBRecord(city,state,zip);

                    database.write(keyArray,record);
            
                    // we wait
                    Thread.yield();
    
                    // and wait again
                    try
                    {
                        Thread.sleep(20);
                    }
                    catch (InterruptedException e)
                    {
                        // stop running on interruption
                        m_active = false;
                    }
                }
            }
            catch (FileNotFoundException not_found)
            {
                addMessage("ERROR: file " + (m_chkSmallDB.isSelected() ? INPUT_80000_NAME : INPUT_ALL_NAME) + " was not found");
                m_active = false;
            }
            catch (Exception ex)
            {
                JOptionPane.showMessageDialog(m_this,"Unable to open or convert those files");
                ex.printStackTrace();
                m_active = false;
            }
            
            // mark thread for disposal
            m_thread = null;  
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessage(">>> DONE <<<");
        }
    }
        
    private class ActionBuild extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // create and start thread
            m_thread = new BuildThread();
            m_thread.start();
        }
    }
    
    private class ExportThread extends Thread
    {
        public void run()
        {
            // now we're running                
            m_active = true;
        
            try
            {
                // open an output text file
                String filename = m_txtExport.getText();
                
                // if there is no file name, use a default
                if (filename.length() < 1)
                {
                    filename = new String("zipexport.csv");
                    m_txtExport.setText(filename);
                }
                
                PrintWriter csvFile = new PrintWriter(new FileOutputStream(filename));
                
                // disable UI while working
                setUIEnabled(false);

                // clear output display
                setMessage("");

                // open database                
                IndexedObjectDatabase database = new IndexedObjectDatabase(DATABASE_NAME,false);
                BTreeIndex codeIndex = new BTreeIndex(ZIP_INDEX_NAME);
                BTreeIndex cityIndex = new BTreeIndex(CITY_INDEX_NAME);
                database.attachIndex(codeIndex);
                database.attachIndex(cityIndex);
                
                // create an iterator
                //BTreeIterator iterator = new BTreeIterator(codeIndex);
                BTreeIterator iterator = new BTreeIterator(cityIndex);
                
                // while we have data to read...
                int counter = 0;
                
                while (m_active)
                {
                    Vector records = database.read(iterator);
                    
                    for (int n = 0; n < records.size(); ++n)
                    {
                        DBRecord rec = (DBRecord)(records.get(n));
                    
                        if (rec != null)
                        {
                            //IntKey key = (IntKey)iterator.getKey();
                            StringKey key = (StringKey)iterator.getKey();
                            csvFile.println(key + "," + rec.m_city + "," + rec.m_state);
                            csvFile.flush();
                            ++counter;
                            addMessage(counter + ": " + rec.m_city + ", " + rec.m_state + " (" + key + ")");
                        }
                    }
                    
                    m_active = iterator.moveNext();

                    // we wait
                    Thread.yield();
    
                    // and wait again
                    try
                    {
                        Thread.sleep(20);
                    }
                    catch (InterruptedException e)
                    {
                        // stop running on interruption
                        m_active = false;
                    }
                }
            }
            catch (Exception ex)
            {
                addMessage("Something broke");
                ex.printStackTrace();
                m_active = false;
            }
            
            // mark thread for disposal
            m_thread = null;  
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessage(">>> DONE <<<");
        }
    }
        
    private class ActionExport extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // create and start thread
            m_thread = new ExportThread();
            m_thread.start();
        }
    }
    
    private class ShowThread extends Thread
    {
        public void run()
        {
            // now we're running                
            m_active = true;
        
            try
            {
                // get from and to keys
                int from = Integer.valueOf(m_txtShowFrom.getText()).intValue();
                int to   = Integer.valueOf(m_txtShowTo.getText()).intValue();
                
                if (from <     0) from = 0;
                if (from > 99999) from = 99999;
                if (to   <     0) to   = 0;
                if (to   > 99999) to   = 99999;
                
                if (from > to)
                {
                    int temp = from;
                    from = to;
                    to = temp;
                }

                m_txtShowFrom.setText("" + from);               
                m_txtShowTo.setText("" + to);               
                
                // disable UI while working
                setUIEnabled(false);
                
                // open database                
                IndexedObjectDatabase database = new IndexedObjectDatabase(DATABASE_NAME,false);
                BTreeIndex codeIndex = new BTreeIndex(ZIP_INDEX_NAME);
                BTreeIndex cityIndex = new BTreeIndex(CITY_INDEX_NAME);
                database.attachIndex(codeIndex);
                database.attachIndex(cityIndex);
                
                // FORWARD                 
                setMessage("Showing from " + from + " to " + to);
                
                // create an iterator and set it to the first object
                BTreeIterator iterator = new BTreeIterator(codeIndex);
                
                if (!iterator.moveTo(new IntKey(from),true))
                {
                    addMessage("'From' ZIP not found.");
                    m_active = false;
                }
                
                // while we have data to read...
                while (m_active)
                {
                    Vector records = database.read(iterator);
                    
                    for (int n = 0; n < records.size(); ++n)
                    {
                        DBRecord rec = (DBRecord)(records.get(n));
                        
                        if (rec != null)
                        {
                            if (rec.m_zip > to)
                            {    
                                m_active = false;
                                break;
                            }
                            else
                                addMessage("found: " + rec.m_zip + "," + rec.m_city + "," + rec.m_state);
                        }
                    }
                    
                    if (m_active)
                    {    
                        m_active = iterator.moveNext();
    
                        // we wait
                        Thread.yield();
        
                        // and wait again
                        try
                        {
                            Thread.sleep(20);
                        }
                        catch (InterruptedException e)
                        {
                            // stop running on interruption
                            m_active = false;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                addMessage("Something broke");
                ex.printStackTrace();
                m_active = false;
            }
            
            // mark thread for disposal
            m_thread = null;  
            
            // re-enable UI while working
            setUIEnabled(true);
            addMessage(">>> DONE <<<");
        }
    }
        
    private class ActionShow extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // create and start thread
            m_thread = new ShowThread();
            m_thread.start();
        }
    }
    
    private class ActionZipSearch extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // disable UI while working
            setUIEnabled(false);
            
            // clear output display
            setMessage("");
            
            try
            {
                // open database
                IndexedObjectDatabase database = new IndexedObjectDatabase(DATABASE_NAME,false);
                BTreeIndex codeIndex = new BTreeIndex(ZIP_INDEX_NAME);
                BTreeIndex cityIndex = new BTreeIndex(CITY_INDEX_NAME);
                database.attachIndex(codeIndex);
                database.attachIndex(cityIndex);
                
                // show what we're looking for                
                addMessage("Searching on key: " + m_txtZip.getText());
                
                // Create a key object
                IntKey key = new IntKey(Integer.valueOf(m_txtZip.getText()).intValue());
                Vector records = database.read(key,codeIndex);
                
                for (int n = 0; n < records.size(); ++n)
                {
                    DBRecord rec = (DBRecord)(records.get(n));
                    addMessage("found: " + rec.m_zip + "," + rec.m_city + "," + rec.m_state);
                }
            }
            catch (Exception ex)
            {
                addMessage("Error finding city record from ZIP code");
            }
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessage(">>> DONE <<<");
        }
    }
    
    private class ActionCitySearch extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // disable UI while working
            setUIEnabled(false);
            
            // clear output display
            setMessage("");
            
            try
            {
                // open database
                IndexedObjectDatabase database = new IndexedObjectDatabase(DATABASE_NAME,false);
                BTreeIndex codeIndex = new BTreeIndex(ZIP_INDEX_NAME);
                BTreeIndex cityIndex = new BTreeIndex(CITY_INDEX_NAME);
                database.attachIndex(codeIndex);
                database.attachIndex(cityIndex);
                
                // show what we're looking for                
                addMessage("Searching on key: " + m_txtCity.getText() + "," + m_txtState.getText());
                
                // Create a key object
                StringKey key = new StringKey(m_txtCity.getText().trim().toUpperCase() + m_txtState.getText().trim().toUpperCase());
                Vector records = database.read(key,cityIndex);
                
                for (int n = 0; n < records.size(); ++n)
                {
                    DBRecord rec = (DBRecord)(records.get(n));
                    addMessage("found: " + rec.m_zip + "," + rec.m_city + "," + rec.m_state);
                }
            }
            catch (Exception ex)
            {
                addMessage("Error finding city record from city and state code " + m_txtCity.getText().trim().toUpperCase() + m_txtState.getText().trim().toUpperCase());
            }
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessage(">>> DONE <<<");
        }
    }
    
    private class ActionDelete extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // disable UI while working
            setUIEnabled(false);
            
            // clear output display
            setMessage("");
            
            try
            {
                // open database
                IndexedObjectDatabase database = new IndexedObjectDatabase(DATABASE_NAME,false);
                BTreeIndex codeIndex = new BTreeIndex(ZIP_INDEX_NAME);
                BTreeIndex cityIndex = new BTreeIndex(CITY_INDEX_NAME);
                database.attachIndex(codeIndex);
                database.attachIndex(cityIndex);
                
                // show what we're looking for                
                addMessage("Deleting key: " + m_txtDelete.getText());
                
                // Create a key object
                IntKey key = new IntKey(Integer.valueOf(m_txtDelete.getText()).intValue());
                database.remove(key,codeIndex);
            }
            catch (KeyNotFoundException ex)
            {
                addMessage("Given ZIP code does not exist in the database");
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
                addMessage("Error finding city record from ZIP code");
            }
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessage(">>> DONE <<<");
        }
    }
    
    private class ActionViewLicense extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            m_output.setText(Jisp.LICENSE);
            m_output.setCaretPosition(0);
        }
    }
    
    private class ActionViewAbout extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            m_output.setText(Jisp.ABOUT);
            m_output.setCaretPosition(0);
        }
    }
    
    // application working storage
    private JFrame      m_frame;    
    private JispDemo    m_this   = this;
    private ClassLoader m_loader;
    private String      m_cwd;

    // threading tools    
    private Thread  m_thread = null;
    private boolean m_active = false;

    // main tabbed window display
    private JPanel        m_mainPanel;
    private JTabbedPane   m_actionTabs;
    
    // output display area
    private JTextArea  m_output;
    private int        m_outputCount;
    private static final int OUTPUT_LIMIT = 100;
    
    // about controls
    private JButton    m_btnLicense;
    private JButton    m_btnAbout;
    
    // build DB controls
    private JButton    m_btnBuild;
    private JTextField m_txtOrder;
    private JCheckBox  m_chkSmallDB;
    private JButton    m_btnCancelBuild;

    // export to text file
    private JButton    m_btnExport;
    private JTextField m_txtExport;
    private JButton    m_btnCancelExport;
    
    // show sequence of records
    private JButton    m_btnShow;
    private JTextField m_txtShowFrom;
    private JTextField m_txtShowTo;
    
    // ZIP search controls
    private JButton    m_btnZipSearch;
    private JTextField m_txtZip;
    
    // City search controls
    private JButton    m_btnCitySearch;
    private JTextField m_txtCity;
    private JTextField m_txtState;
    
    // ZIP delete controls
    private JButton    m_btnDelete;
    private JTextField m_txtDelete;
    
    // set UI state
    private void setUIEnabled(boolean enabled)
    {
        m_btnBuild.setEnabled(enabled);
        m_txtOrder.setEnabled(enabled);
        m_chkSmallDB.setEnabled(enabled);
        m_btnCancelBuild.setEnabled(!enabled);
        m_btnExport.setEnabled(enabled);
        m_txtExport.setEnabled(enabled);
        m_btnCancelExport.setEnabled(!enabled);
        m_btnShow.setEnabled(enabled);
        m_txtShowFrom.setEnabled(enabled);
        m_txtShowTo.setEnabled(enabled);
        m_btnZipSearch.setEnabled(enabled);
        m_txtZip.setEnabled(enabled);
        m_btnDelete.setEnabled(enabled);
        m_txtDelete.setEnabled(enabled);
        m_output.setEnabled(enabled);
        m_btnLicense.setEnabled(enabled);
        m_btnAbout.setEnabled(enabled);
    }
    
    // add a line of text to the output display
    private void setMessage(String text)
    {
        m_outputCount = 1;
        m_output.setText(text);
    }
    
    // add a line of text to the output display
    private void addMessage(String text)
    {
        ++m_outputCount;
        String s = m_output.getText();
        
        if (m_outputCount > OUTPUT_LIMIT)
        {
            int n = s.indexOf('\n',1);
            
            if (n > 0)
                s = s.substring(n);
        }

        m_output.setText(s + "\n" + text);
        m_output.setCaretPosition(m_output.getText().length());
    }
    
    // constructor
    public JispDemo(JFrame owner)
    {
        // base class construction
        super();
        
        // remember owning frame
        m_frame = owner;
        
        // get a class loader for loading resources
        m_loader = ResourceAnchor.class.getClassLoader();
        
        // load look-and-feel
        try
        {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        }
        catch (Exception ex)
        {
            System.err.println("Error loading look-and-feel:\n" + ex);
        }
        
        // get the current directory
        m_cwd = System.getProperty("user.dir");
        
        // set borders and layout for main display
        setLayout(new BorderLayout());
        
        // create another panel to hold search bar
        m_mainPanel = new JPanel();
        m_mainPanel.setLayout(new BorderLayout());
        
        // create the world pane
        m_actionTabs = new JTabbedPane(); // JTabbedPane.LEFT);

        // build DB controls
        m_btnBuild = new JButton("Build Database");
        m_btnBuild.addActionListener(new ActionBuild());
        m_txtOrder = new JTextField(4);
        m_txtOrder.setText("64");
        m_chkSmallDB = new JCheckBox("Use small database");
        m_chkSmallDB.setSelected(true);
        m_btnCancelBuild = new JButton("Cancel");
        m_btnCancelBuild.addActionListener(new ActionCancel());
        m_btnCancelBuild.setEnabled(false);
        
        JPanel buildPanel = new JPanel();
        buildPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        buildPanel.add(m_btnBuild);
        buildPanel.add(new JLabel(" Order: ",JLabel.RIGHT));
        buildPanel.add(m_txtOrder);
        buildPanel.add(m_chkSmallDB);
        buildPanel.add(m_btnCancelBuild);
        
        // export controls
        m_btnExport = new JButton("Export to CSV");
        m_btnExport.addActionListener(new ActionExport());
        m_txtExport = new JTextField(20);
        m_txtExport.setText("zipexport.csv");
        m_btnCancelExport = new JButton("Cancel");
        m_btnCancelExport.addActionListener(new ActionCancel());
        m_btnCancelExport.setEnabled(false);
        
        JPanel exportPanel = new JPanel();
        exportPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        exportPanel.add(m_btnExport);
        exportPanel.add(new JLabel(" File: ",JLabel.RIGHT));
        exportPanel.add(m_txtExport);
        exportPanel.add(m_btnCancelExport);
        
        // show controls
        m_btnShow = new JButton("Show");
        m_btnShow.addActionListener(new ActionShow());
        m_txtShowFrom = new JTextField(5);
        m_txtShowFrom.setText("80201");
        m_txtShowTo = new JTextField(5);
        m_txtShowTo.setText("80299");
        
        JPanel showPanel = new JPanel();
        showPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        showPanel.add(m_btnShow);
        showPanel.add(new JLabel(" From ",JLabel.RIGHT));
        showPanel.add(m_txtShowFrom);
        showPanel.add(new JLabel(" to ",JLabel.RIGHT));
        showPanel.add(m_txtShowTo);

        // search for ZIP
        m_btnZipSearch = new JButton("Find City for ZIP");
        m_btnZipSearch.addActionListener(new ActionZipSearch());
        m_txtZip = new JTextField(5);

        JPanel zipPanel = new JPanel();
        zipPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        zipPanel.add(m_btnZipSearch);
        zipPanel.add(new JLabel(" ZIP code: ", JLabel.RIGHT));
        zipPanel.add(m_txtZip);

        // search for City
        m_btnCitySearch = new JButton("Find ZIP by City");
        m_btnCitySearch.addActionListener(new ActionCitySearch());
        m_txtCity  = new JTextField(24);
        m_txtCity.setText("Denver");
        m_txtState = new JTextField(3);
        m_txtState.setText("CO");
        
        JPanel cityPanel = new JPanel();
        cityPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        cityPanel.add(m_btnCitySearch);
        cityPanel.add(new JLabel(" City: ", JLabel.RIGHT));
        cityPanel.add(m_txtCity);
        cityPanel.add(new JLabel(" State: ", JLabel.RIGHT));
        cityPanel.add(m_txtState);
        
        // delete a ZIP code
        m_btnDelete = new JButton("Delete a ZIP Code");
        m_btnDelete.addActionListener(new ActionDelete());
        m_txtDelete = new JTextField(10);

        JPanel deletePanel = new JPanel();
        deletePanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        deletePanel.add(m_btnDelete);
        deletePanel.add(new JLabel(" ZIP code: ", JLabel.RIGHT));
        deletePanel.add(m_txtDelete);

        // about panel
        JLabel logo = new JLabel();
        URL iconUrl = ResourceAnchor.class.getClassLoader().getResource("CoyoteGulchLogo.png");
        logo.setIcon(new ImageIcon(iconUrl));
        m_btnLicense = new JButton("License");
        m_btnLicense.addActionListener(new ActionViewLicense());
        m_btnAbout = new JButton("About");
        m_btnAbout.addActionListener(new ActionViewAbout());
        
        JPanel aboutPanel = new JPanel();
        aboutPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        aboutPanel.add(logo);
        aboutPanel.add(new JLabel("Jisp " + Jisp.VERSION + " -- the Java Indexed Serialization Package"));
        aboutPanel.add(m_btnLicense);
        aboutPanel.add(m_btnAbout);
        
        // create control panel
        m_actionTabs.add("About",aboutPanel);
        m_actionTabs.add("Build",buildPanel);
        m_actionTabs.add("Export",exportPanel);
        m_actionTabs.add("Display",showPanel);
        m_actionTabs.add("Search by ZIP",zipPanel);
        m_actionTabs.add("Search by City",cityPanel);
        m_actionTabs.add("Delete",deletePanel);
        
        // create output display area
        m_output = new JTextArea();
        m_output.setEditable(false);
        m_output.setText(Jisp.ABOUT);
        
        JScrollPane outputPanel = new JScrollPane();
        outputPanel.getViewport().add(m_output);

        // add the whole thing to the work panel
        m_mainPanel.add(m_actionTabs, BorderLayout.NORTH);
        m_mainPanel.add(outputPanel, BorderLayout.CENTER);
        
        // add main panel to display
        add(new JLabel(" "), BorderLayout.NORTH);
        add(m_mainPanel, BorderLayout.CENTER);
    }
    
    // main entry point
    public static void main(String[] args)
    {
        try
        {
            String javaVersion = System.getProperty("java.version");
            
            if (javaVersion.compareTo("1.3.0") < 0)
            {
                System.err.println("JispDemo requires a version 1.3.0 or higher JVM");
                System.exit(1);
            }
             
            JFrame frame = new JFrame();
            frame.setTitle(Preferences.getAppName() + " " + Preferences.getAppVersion());
            frame.getContentPane().setLayout(new BorderLayout());
            frame.getContentPane().add(new JispDemo(frame),BorderLayout.CENTER);
            frame.addWindowListener(new AppCloser());
            frame.pack();
            frame.setSize(720,540);
            
            // center the frame
            Rectangle screenRect = frame.getGraphicsConfiguration().getBounds();
            int centerWidth = screenRect.width < frame.getSize().width ? screenRect.x : screenRect.x + screenRect.width/2 - frame.getSize().width/2;
            int centerHeight = screenRect.height < frame.getSize().height ? screenRect.y : screenRect.y + screenRect.height/2 - frame.getSize().height/2;
            frame.setLocation(centerWidth, centerHeight);
            
            frame.setVisible(true);
        }
        catch (Throwable t)
        {
            System.err.println("Uncaught exception: " + t);
            t.printStackTrace();
        }
    }
}
