//-----------------------------------------------------------------------
//  Jisp - The Java Indexed Serialization Package
//  Version 3
//
//  StressBTree.java
//-----------------------------------------------------------------------
// 
//  Copyright (C) 1996-2003 Scott Robert Ladd
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or (at
//  your option) any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//-----------------------------------------------------------------------
//  A COMMERCIAL LICENSE IS AVAILABLE from the author; you can contact
//  him at his web site or via the following e-mail address:
//
//      coyote@coyotegulch.com
//
//  See the LICENSE.POLICY file for more information on Jisp is licensed.
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//        http://www.coyotegulch.com
//-----------------------------------------------------------------------

import java.io.*;
import java.util.*;
import com.coyotegulch.jisp.*;

/**
 * <code>StressBTree</code> performs multiple random insert and remove options on a BTreeIndex,
 * testing the index algorithm and integrity using external checks.
 * <p>
 * @author  Scott Robert Ladd
 * @see     com.coyotegulch.jisp.BTreeIndex
 * @see     com.coyotegulch.jisp.BTreeIterator
 */
public class StressBTree
{
    // build a database, in a separate thread    
    private static final String BTREE_NAME = "StressBTree.btree";

    // function to display error information
    static private void logical_fallacy(String message, int iteration, int test_size, long seed, IntKey key, int max_key)
    {
        System.out.flush();
    
        System.err.println("\n\nERROR: "  + message
                         + "\niteration " + iteration + " of " + test_size
                         + "\nmax key = " + max_key
                         + "\n   seed = " + seed
                         + "\n    key = " + key.intValue() + "\n");
        
        System.exit(1);
    }

    // main entry point
    public static void main(String[] args)
    {
        try
        {
            // check Java version
            String javaVersion = System.getProperty("java.version");

            if (javaVersion.compareTo("1.4.0") < 0)
            {
                System.err.println("StressBTree requires a version 1.4.0 or higher JVM");
                System.exit(1);
            }

            // parameters for this run
            long value    =      -1L;
            int test_size = 1000000;
            int max_key   =    1024;
            int order     =       4;
            long seed     = System.currentTimeMillis();

            // modify with any command-line arguments
            // this is pretty simple, but nothing more is really necessary
            if (args.length > 0)
                test_size = Integer.valueOf(args[0]).intValue();

            if (args.length > 1)
                max_key = Integer.valueOf(args[1]).intValue();

            if (args.length > 2)
                order = Integer.valueOf(args[2]).intValue();

            if (args.length > 3)
                seed = Long.valueOf(args[3]).longValue();

            
            // display a heading
            System.out.println("\nStressBTree -- a test for the BTreeIndex and BTreeIterator objects\n");
            System.out.println("testing with order = " + order);
            System.out.println("         test size = " + test_size);
            System.out.println("           max key = " + max_key);
            System.out.println("              seed = " + seed);
                            
            // delete old files
            System.out.println("deleting old files");
            File killit = new File(BTREE_NAME);
            
            if (killit.exists())
                killit.delete();
                
            // create database
            System.out.println("creating and initializing index");
            BTreeIndex index = new BTreeIndex(BTREE_NAME,order,false);
            
            // initialize random number generator
            System.out.println("creating random number generator");
            Random randgen = new Random(seed);

            // list of keys already in tree
            boolean key_flags[] = new boolean[max_key];
            
            // flag to break out of loops if an exception is thrown
            boolean broken = false;

            // timing loop    
            System.out.println("starting timing loop");
            long startTime = System.currentTimeMillis();
        
            // working storage
            int n;
            IntKey key;

            // main test 
            System.out.print("insertion/deletion test\n");
            for (n = 0; n < test_size; ++ n)
            {
                // clear the page cache every thousand operations
                if ((n % 1000) == 0)
                    index.emptyPageCache();
                
                // pick random key
                key = new IntKey(randgen.nextInt(max_key));

                // insert key
                System.out.print(n + ": inserting " + key.intValue() + "; ");
                
                // flag for trapping duplicate keys
                boolean is_dupe = false;
                
                try
                {
                    index.insertKey(key,(long)(key.intValue()));
                    // index.dumpTree(System.out);
                }
                catch (DuplicateKeyException dupe)
                {
                    is_dupe = true;
                }
                
                if (is_dupe)
                {
                    if (key_flags[key.intValue()])
                    {
                        System.out.print("rejected, key already in btree; ");

                        try
                        {
                            index.removeKeyAll(key);
                            System.out.println("removed");
                            key_flags[key.intValue()] = false;
                            // index.dumpTree(System.out);
                        }
                        catch (Throwable t)
                        {
                            logical_fallacy("could not remove",n,test_size,seed,key,max_key);
                        }
                    }
                    else
                        logical_fallacy("stored duplicate key",n,test_size,seed,key,max_key);
                }
                else
                {
                    if (key_flags[key.intValue()])
                        logical_fallacy("should have hit a duplicate",n,test_size,seed,key,max_key);
                    else
                    {
                        System.out.println("succeeded");
                        key_flags[key.intValue()] = true;
                    }
                }
            }
            
            // verify via iterator
            boolean test_flags[] = new boolean[max_key];
            
            BTreeIterator iterator = new BTreeIterator(index);

            // while we have data to read...
            while (iterator.isValid())
            {
                // System.out.println("found key " + iterator.getKey() + " with value " + iterator.getRecPtr());
                test_flags[((IntKey)iterator.getKey()).intValue()] = true;
                if (!iterator.moveNext())
                    break;
            }
            
            for (n = 0; n < max_key; ++n)
            {
                if (key_flags[n] != test_flags[n])
                    System.out.println("ERROR: key_flags[" + n + "] = " + key_flags[n]
                                         + ", test_flags[" + n + "] = " + test_flags[n]);
            }

            // make sure; check internal values match btree
            System.out.print("verifying\n");
            for (n = 0; n < max_key; ++n)
            {
                key = new IntKey(n);
                System.out.print("finding key " + key.intValue());
                try
                {
                    value = index.findKey(key)[0];
                }
                catch (KeyNotFoundException knf)
                {
                    if (key_flags[n])
                        logical_fallacy("key not found, and should have been",n,test_size,seed,key,max_key);
                    else
                    {    
                        System.out.println(" not found, as expected");
                        continue;
                    }
                }
            
                if (key_flags[n])
                {
                    System.out.print(" found, ");

                    if (value == (long)n)
                        System.out.println("good data");
                    else
                        logical_fallacy("bad data",n,test_size,seed,key,max_key);
                }
                else
                    logical_fallacy("key found, and should not have been",n,test_size,seed,key,max_key);
            }

            // print final run time
            System.out.println("\ndone\n\ntest time: " + (System.currentTimeMillis() - startTime));
        }
        catch (Throwable t)
        {
            t.printStackTrace();
        }
    }
}

